/* This file was generated by syncqt. */
#ifndef QT_QTREPPARSER_VERSION_H
#define QT_QTREPPARSER_VERSION_H

#define QTREPPARSER_VERSION_STR "5.12.8"

#define QTREPPARSER_VERSION 0x050C08

#endif // QT_QTREPPARSER_VERSION_H
