/****************************************************************************
**
** Copyright (C) 2017 Ford Motor Company
** Contact: https://www.qt.io/licensing/
**
** This file is part of the QtRemoteObjects module of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:GPL-EXCEPT$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 as published by the Free Software
** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
** $QT_END_LICENSE$
**
****************************************************************************/

#ifndef TESTS_ENGINE_H
#define TESTS_ENGINE_H

#include "rep_engine_source.h"

class Engine : public EngineSimpleSource
{
    Q_OBJECT
    Q_PROPERTY(bool purchasedPart READ purchasedPart WRITE setpurchasedPart)

public:
    Engine(int cylinders = 4, QObject *parent = nullptr);
    ~Engine() override;

    bool start() override;
    void increaseRpm(int deltaRpm) override;

    void unnormalizedSignature(int, int) override {}

    Temperature temperature() override;
    void setTemperature(const Temperature &value);

    void setSharedTemperature(const Temperature::Ptr &) override {}

    bool purchasedPart() {return _purchasedPart;}

public Q_SLOTS:
    void setpurchasedPart(bool value);

    QString myTestString() override { return _myTestString; }
    void setMyTestString(QString value) override { _myTestString = value; }

private:
    bool _purchasedPart;
    QString _myTestString;
    Temperature _temperature;
};

#endif
