const MyPOD initialValue(42, 3.14, QStringLiteral("SubClass"));
const MyPOD updatedValue(-1, 123.456, QStringLiteral("Updated"));
const VariantPOD podValue(10, 11);
const int initialI = 100;
const int updatedI = 200;

